/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include <stdio.h>
#include <libmbtk/global_typedefs.h>
#include <libmbtk/handler.h>
#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/widget.h>
#include <libmbtk/widget_imp.h>

#include "hand_hovradio.h"

static const char my_name[] = "hand_hovradio";

void mbtk_hand_hovradio_activated(mbtk_widget_t *widget)
{
	mbtk_hand_radio_activated(widget);
}

void mbtk_hand_hovradio_deactivate(mbtk_handler_hovradio_t *hr, int fire_cb)
{
	mbtk_hand_radio_deactivate((mbtk_handler_radio_t *)hr, fire_cb);
}


void mbtk_hand_hovradio_invoke(mbtk_widget_t *widget)
{
	mbtk_kw_t kw_invoked, kw_child_invocation_finished;

	mbtk_kw_cache(kw_invoked, "invoked");
	if (mbtk_callback_noarg(widget, kw_invoked) == MBTK_EVENT_HANDLED) {
		mbtk_widget_t *parent = (mbtk_widget_t *)widget->hvbox.parent;
		/* tell parent about the succesful invocation; menu will close using this */
		mbtk_kw_cache(kw_child_invocation_finished, "child_invocation_finished");
		if (parent != NULL)
			mbtk_callback_noarg(parent, kw_child_invocation_finished);
	}
}

void mbtk_hand_hovradio_cancel(mbtk_widget_t *widget)
{
	mbtk_kw_t kw_cancel;
	mbtk_widget_t *parent = (mbtk_widget_t *)widget->hvbox.parent;

	/* tell parent about the cancel; menu will close using this */
	mbtk_kw_cache(kw_cancel, "cancel");
	if (parent != NULL)
		mbtk_callback_noarg(parent, kw_cancel);
}


static mbtk_event_handled_t hovradio_event(mbtk_widget_t *widget, mbtk_event_t *ev)
{
	mbtk_handler_radio_t *hr = widget->handler_data;

	assert(widget->handler->name == my_name);

	switch(ev->type) {
		case MBTK_EV_KEY_PRESS:
			if (ev->data.key.sym == MBTK_KS_LEFT) {
			}
			else if (ev->data.key.sym == MBTK_KS_RIGHT) {
			}
			else if (ev->data.key.sym == MBTK_KS_ENTER) {
				mbtk_hand_hovradio_invoke(widget);
			}
			else if (ev->data.key.sym == MBTK_KS_ESC) {
				mbtk_hand_hovradio_cancel(widget);
			}
			break;
		case MBTK_EV_MOUSE_PRESS:
			if (ev->data.mouse.button == MBTK_MB_RIGHT) {
				mbtk_hand_hovradio_cancel(widget);
				break;
			}
			if (ev->data.mouse.button == MBTK_MB_LEFT) {
				if (!widget->flags.activated) { /* first click, just activate */
					mbtk_hand_radio_activated(widget);
					mbtk_window_set_focus(widget->win, widget, 1);
				}
				else
					mbtk_hand_hovradio_invoke(widget); /* clicked on an already activated item: invoke! */

				return MBTK_EVENT_HANDLED;
			}
			break;

		case MBTK_EV_MOUSE_MOTION:
			if ((hr->active >= 0) && !widget->flags.activated) {
				mbtk_hand_radio_activated(widget);
				mbtk_window_set_focus(widget->win, widget, 1);
				return MBTK_EVENT_HANDLED;
			}
			break;

		case MBTK_EV_FOCUS_MOUSE_PRESS_AWAY:
			if (widget->flags.activated) {
				mbtk_kw_t kw_toggled;

				mbtk_kw_cache(kw_toggled, "toggled");

				mbtk_widget_set_activated(widget, 0);
				mbtk_callback_noarg(widget, kw_toggled);
			}

			hr->active = -1;
			break;

		default:;
	}
	return MBTK_EVENT_NOT_HANDLED;
}

static void hovradio_widget_remove(mbtk_widget_t *widget)
{
	mbtk_handler_radio_t *hr = widget->handler_data;

	assert(widget->handler->name == my_name);

	mbtk_handler_hovradio_del_widget(hr, widget);
}

static void hovradio_handler_destroy(mbtk_handler_t *hr)
{
	mbtk_handler_radio_free((mbtk_handler_radio_t *)hr);
}



void mbtk_handler_hovradio_init(mbtk_window_t *parent, mbtk_handler_radio_t *hr)
{
	memset(hr, 0, sizeof(mbtk_handler_hovradio_t));

	hr->hdr.name = my_name;
	hr->hdr.event = hovradio_event;
	hr->hdr.widget_remove = hovradio_widget_remove;
	hr->hdr.handler_destroy = hovradio_handler_destroy;

	hr->active = -1;

	mbtk_window_add_handler(parent, &hr->hdr);
}


mbtk_handler_hovradio_t *mbtk_handler_hovradio_alloc(mbtk_window_t *parent)
{
	mbtk_handler_radio_t *hr;

	hr = malloc(sizeof(mbtk_handler_hovradio_t));
	mbtk_handler_hovradio_init(parent, hr);
	hr->alloced = 1;
	return hr;
}

void mbtk_handler_hovradio_uninit(mbtk_handler_hovradio_t *hr)
{
	mbtk_handler_radio_uninit(hr);
}

void mbtk_handler_hovradio_free(mbtk_handler_hovradio_t *hr)
{
	mbtk_handler_radio_free(hr);
}


void mbtk_handler_hovradio_add_widget(mbtk_handler_hovradio_t *hr, mbtk_widget_t *w)
{
	mbtk_handler_radio_add_widget(hr, w);
}

void mbtk_handler_hovradio_del_widget(mbtk_handler_hovradio_t *hr, mbtk_widget_t *w)
{
	mbtk_handler_radio_del_widget_nochk(hr, w);
}

