/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/widget.h>
#include <libmbtk/widget_imp.h>
#include <libmbtk/widget/rim.h>
#include <libmbtk/widget/box.h>
#include <libmbtk/widget/label.h>
#include <libmbtk/widget/Common.h>

/*
Design: this widget is a VBOX of two widgets. The top widget is the title,
the bottom wiget is the content (child widget). Normally the rim is drawn
in overlay, across the title widget to make it look like they are combined.
The fallback draw needs to work from boxes so it draws slighlty differently.

The title widget is really a HBOX of three sections:
 - left spacer (empty vbox)
 - the actual title widget supplied by the caller
 - right soacer (empty vbox)

Depending on the alignment, these spacers may span=fill or be some fixed width.
*/

static void mbtk_rim_imp_destroy(mbtk_widget_t *w) { }

int mbtk_rim_fallback_minsize(mbtk_widget_t *w, long *mx, long *my)
{
	long l, r, t, b;
	mbtk_window_t *win = w->win;
	const mbtk_style_t *style = mbtk_widget_style(w);
	mbtk_styleprop_t *sp = mbtk_styleprop_get(w);
	mbtk_widget_flags_t flags = {0};

	flags.activated = 1;
	style->min_box_frame_size(win, sp, flags, &l, &r, &t, &b);

	w->hvbox.fr_l = l*2+0; w->hvbox.fr_r = r*2+2;
	w->hvbox.fr_t = t*2+1; w->hvbox.fr_b = b*2+2;
	*mx = (l+r)*2+2; *my = (t+b)*2+2;

	return 0;
}

void mbtk_rim_fallback_draw(mbtk_widget_t *w)
{
	mbtk_window_t *win = w->win;
	long l, r, t, b;
	const mbtk_style_t *style = mbtk_widget_style(w);
	mbtk_styleprop_t *sp = mbtk_styleprop_get(w);
	mbtk_widget_flags_t flags = {0};
	int titleh2 = w->hvbox.ch_first->h/2; /* first child is the titlebar */
	hvbox_t *title = w->hvbox.ch_first->ch_first->next;

	style->min_box_frame_size(win, sp, flags, &l, &r, &t, &b);

	titleh2 += t+b;

	style->draw_bg_rect(win, sp, flags, w->hvbox.x, w->hvbox.y, w->hvbox.w, w->hvbox.h);
	style->draw_box_frame(win, sp, flags, w->hvbox.x, w->hvbox.y + titleh2, w->hvbox.w, w->hvbox.h-titleh2);
	flags.activated = 1;
	style->draw_box_frame(win, sp, flags, w->hvbox.x+l, w->hvbox.y+t + titleh2, w->hvbox.w-l-r, w->hvbox.h-t-b-titleh2);

	flags.activated = 0;
	style->draw_bg_rect(win, sp, flags, title->x, title->y, title->w, title->h);
}


static const mbtk_widget_imp_t mbtk_rim_imp = {
	mbtk_rim_imp_destroy,

	mbtk_rim_fallback_minsize,
	mbtk_rim_fallback_draw,
	mbtk_comm_frame_fallback_draw_overlay,

	NULL,  /* get_prop */
	NULL,  /* set_prop */

	NULL, /* post_assign */
	NULL, /* pre_unassign */

	NULL, /* pre_appear */
	NULL, /* post_disappear */

	NULL  /* pre_child_unlink */
};

static mbtk_rim_t *mbtk_rim_new_(mbtk_rim_t *caller_alloced, mbtk_widget_t *title)
{
	mbtk_rim_t *rim;
	mbtk_kw_t kw_rim, kw_rim_titlebar;

	mbtk_kw_cache(kw_rim, "rim");
	mbtk_kw_cache(kw_rim_titlebar, "rim.titlebar");

	mbtk_create_wt(rim, kw_rim, &mbtk_rim_imp, caller_alloced);
	rim->w.hvbox.type = HVBOX_VBOX;

	mbtk_box_new(&rim->tbar, 1);
	rim->tbar.w.purpose = kw_rim_titlebar;
	mbtk_box_add_widget((mbtk_box_t *)rim, &rim->tbar.w, 0);


	mbtk_box_new(&rim->tspc1, 0);
	if (title == NULL) {
		mbtk_box_new(&rim->dummy_title, 0);
		mbtk_box_set_minsize(&rim->dummy_title, 40, 10);
		title = &rim->dummy_title.w;
	}
	mbtk_box_new(&rim->tspc2, 0);
	mbtk_box_add_widget(&rim->tbar, &rim->tspc1.w, 0);
	mbtk_box_add_widget(&rim->tbar, title, 0);
	mbtk_box_add_widget(&rim->tbar, &rim->tspc2.w, 0);

	mbtk_widget_embedded(&rim->tbar.w, title);

	mbtk_box_set_minsize(&rim->tspc1, 10, 0);
	mbtk_box_set_minsize(&rim->tspc2, 10, 0);
	mbtk_box_set_span(&rim->tspc2, HVBOX_FILL);

	rim->tbar.no_background = 1;
	rim->tspc1.no_background = 1;
	rim->tspc2.no_background = 1;

	return rim;
}

mbtk_rim_t *mbtk_rim_new(mbtk_rim_t *caller_alloced)
{
	return mbtk_rim_new_(caller_alloced, NULL);
}

mbtk_rim_t *mbtk_rim_new_with_widget(mbtk_rim_t *caller_alloced, mbtk_widget_t *title)
{
	mbtk_rim_t *res = mbtk_rim_new_(NULL, title);
	return res;
}

mbtk_rim_t *mbtk_rim_new_with_label(const char *str)
{
	return mbtk_rim_new_with_widget(NULL, (mbtk_widget_t *)mbtk_label_new(NULL, str));
}


mbtk_widget_t *mbtk_rim_set_child(mbtk_rim_t *rim, mbtk_widget_t *w)
{
	mbtk_widget_t *res, *cnt = (mbtk_widget_t *)rim->w.hvbox.ch_first->next;

	if (cnt == NULL) {
	 /* no second widget in the VBOX means no previous content */
		mbtk_box_add_widget((mbtk_box_t *)rim, w, 0);
		mbtk_widget_embedded((mbtk_widget_t *)w->hvbox.parent, w);
		res = NULL;
	}
	else
		res = mbtk_comm_embed_replace(&rim->w, cnt, w);

	return res;
}

mbtk_widget_t *mbtk_rim_set_title_widget(mbtk_rim_t *rim, mbtk_widget_t *title)
{
	mbtk_widget_t *res, *ttl = (mbtk_widget_t *)rim->w.hvbox.ch_first->ch_first->next;

	res = mbtk_comm_embed_replace(&rim->w, ttl, title);
	if (res == &rim->dummy_title.w) {
		mbtk_widget_destroy(res);
		res = NULL;
	}

	return res;
}

mbtk_widget_t *mbtk_rim_set_title_label(mbtk_rim_t *parent, const char *str)
{
	return mbtk_rim_set_title_widget(parent, (mbtk_widget_t *)mbtk_label_new(NULL, str));
}

mbtk_widget_t *mbtk_rim_subwidget(mbtk_rim_t *rim, mbtk_kw_t subname)
{
	hvbox_t *tbar = rim->w.hvbox.ch_first;
	mbtk_kw_t kw_tbar, kw_tspc1, kw_tspc2, kw_title, kw_content;

	mbtk_kw_cache(kw_tbar, "tbar");
	mbtk_kw_cache(kw_tspc1, "tspc1");
	mbtk_kw_cache(kw_tspc2, "tspc2");
	mbtk_kw_cache(kw_title, "title");
	mbtk_kw_cache(kw_content, "content");

	if (subname == kw_tbar) return (mbtk_widget_t *)tbar;
	if (subname == kw_tspc1) return (mbtk_widget_t *)tbar->ch_first;
	if (subname == kw_title) return (mbtk_widget_t *)tbar->ch_first->next;
	if (subname == kw_tspc2) return (mbtk_widget_t *)tbar->ch_first->next->next;
	if (subname == kw_content) return (mbtk_widget_t *)tbar->next;

	return NULL;
}
