/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/widget.h>
#include <libmbtk/widget_imp.h>
#include <libmbtk/widget/scrolled_wb.h>
#include <libmbtk/widget/textline.h>
#include <libmbtk/widget/box.h>
#include <libmbtk/widget/textedit.h>
#include <libmbtk/widget/Common.h>
#include <libmbtk/handler.h>

/*
Design: this widget is a scolled_wb with a VBOX of textline widgets in main,
one textline per line of text. A pointer to each textline is also stored in
the lines vector so quick by-line-number lookup is possible (e.g. for page-up).
*/

/* How many columns and lines should we aim for in minsize() callback; for
   a single-line textedit MIN_LINES is 1. */
#define MIN_COLS   8
#define MIN_LINES  4

static void mbtk_handler_textedit_install(mbtk_widget_t *widget);

static void mbtk_textedit_imp_destroy(mbtk_widget_t *w)
{
	mbtk_textedit_t *textedit = (mbtk_textedit_t *)w;

	vtp0_uninit(&textedit->lines);
}

static void mbtk_textedit_appear(mbtk_widget_t *w)
{
	mbtk_textedit_t *textedit = (mbtk_textedit_t *)w;

	if (textedit->root_set)
		return;

	textedit->scwb.last.x0 = textedit->scwb.last.y0 = -1; /* trigger update on first draw */
	mbtk_scrolled_set_root_widget(&textedit->scwb.main, &textedit->vbox.w);
	mbtk_comm_scrolled_wb_update(&textedit->scwb);
	textedit->root_set = 1;
}


int mbtk_textedit_fallback_minsize(mbtk_widget_t *w, long *mx, long *my)
{
	mbtk_textedit_t *textedit = (mbtk_textedit_t *)w;
	int wi, asc, dsc, sc_w, sc_h;

	if (mbtk_comm_frame_fallback_minsize(w, mx, my) != 0)
		*mx = *my = 0;

	mbtk_draw_set_font(w->win, w->win->display->default_fid);
	mbtk_draw_query_text(w->win, &wi, &asc, &dsc, "M", 1);

	sc_w = textedit->scwb.scb_vert.w.hvbox.w_min + w->hvbox.spacing + w->hvbox.padding*2 + textedit->scwb.main.w.hvbox.padding*2;
	sc_h = textedit->scwb.scb_horiz.w.hvbox.h_min + w->hvbox.spacing + w->hvbox.padding*2 + textedit->scwb.main.w.hvbox.padding*2;

	*mx += wi*MIN_COLS + sc_w;
	if (!textedit->single_line)
		*my += (asc+dsc+textedit->scwb.main.w.hvbox.spacing+4) * MIN_LINES + sc_h;
	else
		*my += (asc+dsc+4) + sc_h;
	return 0;
}

void mbtk_textedit_fallback_draw_no_frame(mbtk_widget_t *w)
{
	mbtk_textedit_t *textedit = (mbtk_textedit_t *)w;

	/* make sure the vbox fills up scrolled area; this matters if scrolled is
	   larger than the text. This way text background is always drawn over
	   the whole scrolled area and clicks beyond the text on the right are
	   always picked up (clicks are handled from the vbox) */
	mbtk_box_set_minsize(&textedit->vbox, textedit->scwb.main.w.hvbox.w, textedit->scwb.main.w.hvbox.h);

	mbtk_comm_scrolled_wb_update(&textedit->scwb); /* won't do anything if it doesn't need to */
}


void mbtk_textedit_fallback_draw(mbtk_widget_t *w)
{
	mbtk_textedit_fallback_draw_no_frame(w);
	mbtk_comm_frame_fallback_draw(w);
}

static void mbtk_textedit_vbox_pre_child_unlink(mbtk_widget_t *vbox, mbtk_widget_t *child)
{
	mbtk_textedit_t *textedit = vbox->handler_data;
	long n;

	if (textedit->lock_del || textedit->w.destroying)
		return;

	/* remove the corresponding textedit->lines entry */
	for(n = 0; n < textedit->lines.used; n++) {
		if (textedit->lines.array[n] == child) {
			vtp0_remove(&textedit->lines, n, 1);
			if (textedit->cursor.y == n)
				textedit->cursor.y = -1;
			else if (textedit->cursor.y)
				textedit->cursor.y--;
			break;
		}
	}
}

static const mbtk_widget_imp_t mbtk_textedit_imp = {
	mbtk_textedit_imp_destroy,

	mbtk_textedit_fallback_minsize,
	mbtk_textedit_fallback_draw,
	mbtk_comm_frame_fallback_draw_overlay,

	NULL,  /* get_prop */
	NULL,  /* set_prop */

	NULL, /* post_assign */
	NULL, /* pre_unassign */

	mbtk_textedit_appear, /* pre_appear */
	NULL, /* post_disappear */

	NULL  /* pre_child_unlink */
};

static mbtk_widget_imp_t mbtk_textedit_vbox_imp;
static int mbtk_textedit_vbox_imp_inited = 0;

mbtk_textedit_t *mbtk_textedit_new(mbtk_textedit_t *caller_alloced, const char *str, int single_line)
{
	mbtk_textedit_t *textedit;
	mbtk_kw_t kw_textedit;
	mbtk_scrolled_wb_vis_t scbvis = single_line ? MBTK_SCROLLED_WB_NEVER : MBTK_SCROLLED_WB_AUTO;

	mbtk_kw_cache(kw_textedit, "textedit");

	mbtk_create_wt(textedit, kw_textedit, &mbtk_textedit_imp, caller_alloced);

	mbtk_comm_scrolled_wb_new(&textedit->w, &textedit->scwb, scbvis, scbvis);
	mbtk_vbox_new(&textedit->vbox);
	mbtk_box_set_align(&textedit->vbox, HVBOX_LEFT_TOP);
	textedit->root_set = 0;
	textedit->cursor.x = textedit->cursor.y = -1;
	textedit->single_line = single_line;

	mbtk_kw_cache(textedit->scwb.main.w.purpose, "textedit.scrolled");
	mbtk_kw_cache(textedit->vbox.w.purpose, "textedit.vbox");

	vtp0_init(&textedit->lines);

	if ((str != NULL) && (*str != '\0'))
		mbtk_textedit_append_str(textedit, str);

	mbtk_handler_textedit_install(&textedit->vbox.w);
	textedit->vbox.w.handler_data = textedit;

	/* take over vbox implementation so that pre_child_unlink can be handled */
	if (!mbtk_textedit_vbox_imp_inited) {
		memcpy(&mbtk_textedit_vbox_imp, textedit->vbox.w.imp, sizeof(mbtk_widget_imp_t));
		mbtk_textedit_vbox_imp.pre_child_unlink = mbtk_textedit_vbox_pre_child_unlink;
		mbtk_textedit_vbox_imp_inited = 1;
	}
	textedit->vbox.w.imp = &mbtk_textedit_vbox_imp;

	return textedit;
}


int mbtk_textedit_append_line(mbtk_textedit_t *textedit, const char *str, long len)
{
	mbtk_textline_t *lin;

	if (textedit->single_line && (textedit->lines.used != 0))
		return -1;

	lin = mbtk_textline_new(NULL, NULL);
	mbtk_textline_set_str_len(lin, str, len);
	mbtk_box_add_widget(&textedit->vbox, &lin->w, 0);
	vtp0_append(&textedit->lines, lin);
	return 0;
}

int mbtk_textedit_insert_line_after(mbtk_textedit_t *textedit, long after, const char *str, long len)
{
	mbtk_textline_t *existing, *lin;

	if ((after < 0) || (after >= textedit->lines.used) || textedit->single_line)
		return -1;

	existing = textedit->lines.array[after];

	lin = mbtk_textline_new(NULL, NULL);
	mbtk_textline_set_str_len(lin, str, len);
	mbtk_box_add_widget_at(&existing->w, &lin->w, 0);
	vtp0_insert_len(&textedit->lines, after+1, (void **)&lin, 1);

	return 0;
}

void mbtk_textedit_append_str(mbtk_textedit_t *textedit, const char *str)
{
	const char *s, *start;
	long len;

	/* in single line mode: make sure there's a line, append str to it and
	   replace newlines with spaces */
	if (textedit->single_line) {
		long from;
		mbtk_textline_t *line;
		char *s;

		if (textedit->lines.used == 0) {
			mbtk_textedit_append_line(textedit, str, strlen(str));
			line = textedit->lines.array[0];
			from = 0;
		}
		else {
			line = textedit->lines.array[0];
			from = line->str.used;
			gds_append_str(&line->str, str);
			mbtk_textline_str_changed(line);
		}
		for(s = line->str.array+from; *s != '\0'; s++)
			if ((*s == '\r') || (*s == '\n')) *s = ' ';

		return;
	}

	for(start = s = str, len = 0; *s != '\0'; s++,len++) {
		if (*s == '\n') {
			mbtk_textedit_append_line(textedit, start, len);
			start = s+1;
			len = -1;
		}
	}
	if (len > 0)
		mbtk_textedit_append_line(textedit, start, len);
}

int mbtk_textline_del_line(mbtk_textedit_t *textedit, long y)
{
	mbtk_textline_t *line;

	if ((y < 0) || (y >= textedit->lines.used))
		return -1;

	/* remove cursor of cursor is in line */
	if (y == textedit->cursor.y)
		textedit->cursor.y = textedit->cursor.x = -1;

	line = textedit->lines.array[y];

	textedit->lock_del = 1;
	vtp0_remove(&textedit->lines, y, 1);
	mbtk_widget_destroy(&line->w);
	textedit->lock_del = 0;

	return 0;
}

/*** Cursor ***/
long mbtk_textedit_get_page_lines(mbtk_textedit_t *textedit)
{
	hvbox_t *mb = &textedit->scwb.main.w.hvbox;
	hvbox_t *l1 = textedit->vbox.w.hvbox.ch_first;
	long res;

	res = (mb->h - 2 * mb->padding - mb->fr_t - mb->fr_b) / (l1->h + mb->spacing + l1->padding*2);
	if (res > 1)
		res--;
	return res;
}

void mbtk_textedit_auto_scroll(mbtk_textedit_t *textedit)
{
	mbtk_textline_t *line;
	mbtk_window_t *cw = textedit->vbox.w.win;
	long deltax = 0, deltay = 0;

	if ((textedit->cursor.y < 0) || (textedit->cursor.x < 0) || (textedit->cursor.y >= textedit->lines.used))
		return;

	line = textedit->lines.array[textedit->cursor.y];
	if (line == NULL)
		return;

	/* calculate the new size of the text line (line->w.hvbox.w and line->w.hvbox.w) */
	mbtk_win_update_layout(line->w.win, 0);

	if (line->w.hvbox.y + line->w.hvbox.h > cw->content_y0 + cw->h)
		deltay = (line->w.hvbox.y + line->w.hvbox.h) - (cw->content_y0 + cw->h);
	else if (line->w.hvbox.y < cw->content_y0)
		deltay = line->w.hvbox.y - cw->content_y0;

	if (line->cursor_x2 > cw->content_x0 + cw->w)
		deltax = line->cursor_x2 - (cw->content_x0 + cw->w);
	else if (line->cursor_x < cw->content_x0)
		deltax = line->cursor_x - cw->content_x0;

	if ((deltax != 0) || (deltay != 0))
		mbtk_scrolled_set_content_origin(&textedit->scwb.main, cw->content_x0 + deltax, cw->content_y0 + deltay);
}

void mbtk_textedit_cursor_remove(mbtk_textedit_t *textedit)
{
	if ((textedit->cursor.y >= 0) && (textedit->cursor.y < textedit->lines.used)) {
		mbtk_textline_t *oldline = textedit->lines.array[textedit->cursor.y];
		if (oldline != NULL)
			mbtk_textline_cursor_remove(oldline);
		textedit->cursor.x = textedit->cursor.y = -1;
	}
}

int mbtk_textedit_cursor_set_pixel_xy(mbtk_textedit_t *textedit, long x, long y)
{
	long cxp, cy;
	hvbox_t *vb = &textedit->vbox.w.hvbox;
	hvbox_t *l1 = textedit->vbox.w.hvbox.ch_first;
	mbtk_textline_t *line;

	if (l1 == NULL)
		return -1;

	cy  = (y - vb->padding - vb->fr_t) / (l1->h + vb->spacing + l1->padding*2); /* line number */
	cxp = (x - vb->padding - vb->fr_l); /* in pixel */

	if (cy >= textedit->lines.used)
		return -1;

/*	printf("cxp=%ld, cy=%ld    sub=%ld  div=%ld\n", cxp, cy, (long)(vb->padding - vb->fr_t), (long)(l1->h + vb->spacing));*/
	line = textedit->lines.array[cy];
	if (line == NULL)
		return -1;

	if (mbtk_textline_cursor_set_win(line, cxp) != 0)
		return -1;

	if (cy != textedit->cursor.y)
		mbtk_textedit_cursor_remove(textedit);

	textedit->cursor.x = line->cursor;
	textedit->cursor.y = cy;

	if (!textedit->vbox.w.flags.focus)
		mbtk_textline_cursor_remove(line);

	return 0;
}

int mbtk_textedit_cursor_move_left(mbtk_textedit_t *textedit)
{
	if (textedit->cursor.x > 0) {
		mbtk_textline_t *line;

		if ((textedit->cursor.y < 0) || (textedit->cursor.y >= textedit->lines.used))
			return -1;
		line = textedit->lines.array[textedit->cursor.y];
		textedit->cursor.x--;
		if (textedit->vbox.w.flags.focus)
			mbtk_textline_cursor_set_chr(line, textedit->cursor.x, -1);
	}
	else {
		mbtk_textline_t *line;
		long y = textedit->cursor.y - 1;

		if (textedit->cursor.y <= 0) {
			mbtk_textedit_auto_scroll(textedit);
			return 0;
		}

		mbtk_textedit_cursor_remove(textedit);

		textedit->cursor.y = y;
		line = textedit->lines.array[y];
		textedit->cursor.x = line->str.used;
		if (textedit->vbox.w.flags.focus)
			mbtk_textline_cursor_set_chr(line, textedit->cursor.x, -1);
	}

	mbtk_textedit_auto_scroll(textedit);
	return 0;
}

int mbtk_textedit_cursor_move_right(mbtk_textedit_t *textedit)
{
	mbtk_textline_t *line;

	if ((textedit->cursor.y < 0) || (textedit->cursor.y >= textedit->lines.used))
		return -1;

	line = textedit->lines.array[textedit->cursor.y];

	if (textedit->cursor.x < line->str.used) {
		textedit->cursor.x++;
		if (textedit->vbox.w.flags.focus)
			mbtk_textline_cursor_set_chr(line, textedit->cursor.x, -1);
	}
	else {
		long y = textedit->cursor.y + 1;

		if (y >= textedit->lines.used) {
			mbtk_textedit_auto_scroll(textedit);
			return 0;
		}

		mbtk_textedit_cursor_remove(textedit);

		textedit->cursor.y = y;
		line = textedit->lines.array[y];
		textedit->cursor.x = 0;
		if (textedit->vbox.w.flags.focus)
			mbtk_textline_cursor_set_chr(line, textedit->cursor.x, -1);
	}

	mbtk_textedit_auto_scroll(textedit);
	return 0;
}

static int mbtk_textedit_cursor_move_vert_(mbtk_textedit_t *textedit, long dy)
{
	long xpx, y;
	mbtk_textline_t *line, *newline;

	if ((textedit->cursor.y < 0) || (textedit->cursor.y >= textedit->lines.used))
		return -1;

	line = textedit->lines.array[textedit->cursor.y];
	y = textedit->cursor.y + dy;
	if (y < 0) y = 0;
	if (y > textedit->lines.used-1) y = textedit->lines.used-1;
	xpx = line->cursor_x;
	newline = textedit->lines.array[y];

	mbtk_textedit_cursor_remove(textedit);
	mbtk_textline_cursor_set_win(newline, xpx);

	textedit->cursor.x = newline->cursor;
	textedit->cursor.y = y;

	if (!textedit->vbox.w.flags.focus)
		mbtk_textline_cursor_remove(line);

	mbtk_textedit_auto_scroll(textedit);
	return 0;
}

int mbtk_textedit_cursor_move_up(mbtk_textedit_t *textedit)
{
	if (textedit->cursor.y < 0)
		return -1;

	if (textedit->cursor.y == 0)
		return 0;

	return mbtk_textedit_cursor_move_vert_(textedit, -1);
}

int mbtk_textedit_cursor_move_down(mbtk_textedit_t *textedit)
{
	if (textedit->cursor.y < 0)
		return -1;

	if (textedit->cursor.y >= textedit->lines.used-1)
		return 0;

	return mbtk_textedit_cursor_move_vert_(textedit, +1);
}

int mbtk_textedit_cursor_move_page_up(mbtk_textedit_t *textedit)
{
	int res;

	if (textedit->cursor.y < 0)
		return -1;

	if (textedit->cursor.y == 0)
		return 0;

	res = mbtk_textedit_cursor_move_vert_(textedit, -(mbtk_textedit_get_page_lines(textedit)+1));

	if (res == 0) { /* scroll up so cursor is at the bottom */
		mbtk_window_t *cw = textedit->vbox.w.win;
		mbtk_textline_t *line = textedit->lines.array[textedit->cursor.y];
		mbtk_scrolled_set_content_origin(&textedit->scwb.main, cw->content_x0, line->w.hvbox.y-cw->h+line->w.hvbox.h + line->w.hvbox.padding*2);
	}

	return res;
}

int mbtk_textedit_cursor_move_page_down(mbtk_textedit_t *textedit)
{
	int res;

	if (textedit->cursor.y < 0)
		return -1;

	if (textedit->cursor.y >= textedit->lines.used-1)
		return 0;

	res = mbtk_textedit_cursor_move_vert_(textedit, +mbtk_textedit_get_page_lines(textedit));

	if (res == 0) { /* scroll down so cursor is at the bottom */
		mbtk_window_t *cw = textedit->vbox.w.win;
		mbtk_textline_t *line = textedit->lines.array[textedit->cursor.y];
		mbtk_scrolled_set_content_origin(&textedit->scwb.main, cw->content_x0, line->w.hvbox.y-1);
	}

	return res;
}

int mbtk_textedit_cursor_move_home(mbtk_textedit_t *textedit)
{
	mbtk_textline_t *line;

	if ((textedit->cursor.y < 0) || (textedit->cursor.y >= textedit->lines.used))
		return -1;

	line = textedit->lines.array[textedit->cursor.y];
	mbtk_textline_cursor_home(line);
	textedit->cursor.x = line->cursor;

	if (!textedit->vbox.w.flags.focus)
		mbtk_textline_cursor_remove(line);

	mbtk_textedit_auto_scroll(textedit);
	return 0;
}

int mbtk_textedit_cursor_move_end(mbtk_textedit_t *textedit)
{
	mbtk_textline_t *line;

	if ((textedit->cursor.y < 0) || (textedit->cursor.y >= textedit->lines.used))
		return -1;

	line = textedit->lines.array[textedit->cursor.y];
	mbtk_textline_cursor_end(line);
	textedit->cursor.x = line->cursor;

	if (!textedit->vbox.w.flags.focus)
		mbtk_textline_cursor_remove(line);

	mbtk_textedit_auto_scroll(textedit);
	return 0;
}

/*** editing ***/
int mbtk_textedit_add_chr(mbtk_textedit_t *textedit, int chr)
{
	mbtk_textline_t *line;
	int res;
	char tmp[8];

	if ((textedit->cursor.x < 0) || (textedit->cursor.y < 0) || (textedit->cursor.y >= textedit->lines.used))
		return -1;

	line = textedit->lines.array[textedit->cursor.y];
	tmp[0] = chr;
	tmp[1] = '\0';

	res = mbtk_textline_type_str(line, tmp);
	if (res == 0) {
		textedit->cursor.x = line->cursor;
		mbtk_textedit_auto_scroll(textedit);
	}
	return res;
}

int mbtk_textedit_backspace(mbtk_textedit_t *textedit)
{
	mbtk_textline_t *line, *prev;

	/* first char in text, nothing to delete */
	if ((textedit->cursor.x == 0) && (textedit->cursor.y == 0))
		return 0;

	if ((textedit->cursor.x < 0) || (textedit->cursor.y < 0) || (textedit->cursor.y >= textedit->lines.used))
		return -1;

	line = textedit->lines.array[textedit->cursor.y];

	/* backspace at the beginning of a line */
	if (textedit->cursor.x == 0) {
		long cy = textedit->cursor.y, py = textedit->cursor.y - 1;
		prev = textedit->lines.array[py];

		/* move cursor to prev's end */
		textedit->cursor.y = py;
		textedit->cursor.x = prev->str.used;

		if (textedit->vbox.w.flags.focus)
			mbtk_textline_cursor_set_chr(prev, textedit->cursor.x, -1);

		/* edit prev and remove line */
		mbtk_textline_append_str_len(prev, line->str.array, line->str.used);

		mbtk_textline_del_line(textedit, cy);
		mbtk_textedit_auto_scroll(textedit);
		return 0;
	}

	mbtk_textline_backspace(line);
	textedit->cursor.x = line->cursor;
	mbtk_textedit_auto_scroll(textedit);
	return 0;
}

int mbtk_textedit_del(mbtk_textedit_t *textedit)
{
	mbtk_textline_t *line, *next;

	if ((textedit->cursor.x < 0) || (textedit->cursor.y < 0) || (textedit->cursor.y >= textedit->lines.used))
		return -1;

	line = textedit->lines.array[textedit->cursor.y];

	/* last char in text, nothing to delete */
	if ((textedit->cursor.y == textedit->lines.used-1) && (textedit->cursor.x == line->str.used))
		return 0;

	/* del at the end of a line */
	if (textedit->cursor.x == line->str.used) {
		long ny = textedit->cursor.y + 1;
		next = textedit->lines.array[ny];

		/* edit current line and remove next line */
		mbtk_textline_append_str_len(line, next->str.array, next->str.used);

		mbtk_textline_del_line(textedit, ny);
		mbtk_textedit_auto_scroll(textedit);
		return 0;
	}

	mbtk_textline_del(line, 1);
	textedit->cursor.x = line->cursor;
	mbtk_textedit_auto_scroll(textedit);
	return 0;
}

int mbtk_textedit_enter(mbtk_textedit_t *textedit)
{
	mbtk_textline_t *line;

	if ((textedit->cursor.x < 0) || (textedit->cursor.y < 0) || (textedit->cursor.y >= textedit->lines.used))
		return -1;

	line = textedit->lines.array[textedit->cursor.y];

	/* insert a new line and truncate current line */
	if (mbtk_textedit_insert_line_after(textedit, textedit->cursor.y, line->str.array+textedit->cursor.x, line->str.used-textedit->cursor.x) != 0)
		return -1;
	mbtk_textline_truncate(line, textedit->cursor.x);

	mbtk_textline_cursor_remove(line);
	textedit->cursor.x = 0;
	textedit->cursor.y++;

	line = textedit->lines.array[textedit->cursor.y];

	if (textedit->vbox.w.flags.focus)
		mbtk_textline_cursor_set_chr(line, 0, -1);

	mbtk_textedit_auto_scroll(textedit);
	return 0;
}


/*** handler ***/

static const char my_name[] = "hand_textedit";

typedef struct mbtk_textedit_handler_s {
	mbtk_handler_t hdr;
} mbtk_textedit_handler_t;

static mbtk_textedit_handler_t textedit_handler;
static int textedit_handler_inited = 0;

#define DONE return MBTK_EVENT_HANDLED

static mbtk_event_handled_t textedit_event(mbtk_widget_t *widget, mbtk_event_t *ev)
{
	mbtk_textedit_t *textedit = (mbtk_textedit_t *)widget->handler_data;

	if (textedit->single_line && (ev->type == MBTK_EV_KEY_PRESS)) {
		/* ignore these in single line mode */
		switch(ev->data.key.sym) {
			case MBTK_KS_ENTER:
			case MBTK_KS_UP:
			case MBTK_KS_DOWN:
			case MBTK_KS_PGUP:
			case MBTK_KS_PGDN:
				return MBTK_EVENT_NOT_HANDLED;
			default: break;
		}
	}

	switch(ev->type) {
		case MBTK_EV_MOUSE_PRESS:
			if (ev->data.mouse.button == MBTK_MB_LEFT) {
				textedit->focus_no_cursor = 1; /* so that cursor is not drawn in MBTK_EV_FOCUS_RECEIVED */
				mbtk_window_set_focus(widget->win, widget, 1);
				textedit->focus_no_cursor = 0;
				mbtk_textedit_cursor_set_pixel_xy(textedit, ev->data.mouse.x - widget->hvbox.x, ev->data.mouse.y - widget->hvbox.y);
			}
			break;
		case MBTK_EV_KEY_PRESS:
			switch(ev->data.key.sym) {
				case MBTK_KS_BACKSPACE: mbtk_textedit_backspace(textedit); DONE;
				case MBTK_KS_DEL:       mbtk_textedit_del(textedit); DONE;
				case MBTK_KS_ENTER:     mbtk_textedit_enter(textedit); DONE;
				case MBTK_KS_TAB:
					/* TODO: tab doesn't render properly */
					break;
				case MBTK_KS_LEFT:   mbtk_textedit_cursor_move_left(textedit); DONE;
				case MBTK_KS_RIGHT:  mbtk_textedit_cursor_move_right(textedit); DONE;
				case MBTK_KS_UP:     mbtk_textedit_cursor_move_up(textedit); DONE;
				case MBTK_KS_DOWN:   mbtk_textedit_cursor_move_down(textedit); DONE;
				case MBTK_KS_PGUP:   mbtk_textedit_cursor_move_page_up(textedit); DONE;
				case MBTK_KS_PGDN:   mbtk_textedit_cursor_move_page_down(textedit); DONE;
				case MBTK_KS_HOME:   mbtk_textedit_cursor_move_home(textedit); DONE;
				case MBTK_KS_END:    mbtk_textedit_cursor_move_end(textedit); DONE;
				default: break;
			}
			break;
		case MBTK_EV_KEY_RELEASE:
			if (ev->data.key.edit != 0) {
				mbtk_textedit_add_chr(textedit, ev->data.key.edit);
				return MBTK_EVENT_HANDLED;
			}
			break;

		case MBTK_EV_FOCUS_RECEIVED:
			if (!textedit->focus_no_cursor && (textedit->cursor.y >= 0) && (textedit->cursor.y < textedit->lines.used))
				mbtk_textline_cursor_set_chr(textedit->lines.array[textedit->cursor.y], textedit->cursor.x, -1);
			break;

		case MBTK_EV_FOCUS_LOST:
			if ((textedit->cursor.y >= 0) && (textedit->cursor.y < textedit->lines.used))
				mbtk_textline_cursor_remove(textedit->lines.array[textedit->cursor.y]);
			break;

		default: break;
	}

	return MBTK_EVENT_NOT_HANDLED;
}

#undef DONE

static void mbtk_handler_textedit_install(mbtk_widget_t *widget)
{
	if (widget->handler != NULL)
		mbtk_widget_del_handler(widget);

	if (!textedit_handler_inited) {
		textedit_handler.hdr.name = my_name;
		textedit_handler.hdr.event = textedit_event;
		textedit_handler_inited = 1;
	}
	widget->handler = &textedit_handler.hdr;
}

mbtk_widget_t *mbtk_textedit_subwidget(mbtk_textedit_t *textedit, mbtk_kw_t subname)
{
	hvbox_t *ch0 = textedit->w.hvbox.ch_first;
	mbtk_kw_t kw_scrolled, kw_scv, kw_sch;

	mbtk_kw_cache(kw_scrolled, "scrolled");
	mbtk_kw_cache(kw_scv, "srollbar.vert");
	mbtk_kw_cache(kw_sch, "srollbar.horiz");

	if (subname == kw_scrolled) return (mbtk_widget_t *)ch0;
	if (subname == kw_scv) return (mbtk_widget_t *)ch0->next;
	if (subname == kw_sch) return (mbtk_widget_t *)ch0->next->next;

	return NULL;
}
