/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#ifndef LIBMBTK_WIDGET_TEXTEDIT_H
#define LIBMBTK_WIDGET_TEXTEDIT_H

#include <libmbtk/config.h>
#include <libmbtk/widget.h>
#include <libmbtk/widget/scrolled_wb.h>
#include <libmbtk/widget/box.h>
#include <libmbtk/widget_inlines.h>

/* Textedit: single line text entry or multiline text editor. Handler is
   coupled. Composite widget. */

typedef struct mbtk_textedit_s {
	mbtk_widget_t w;
	mbtk_comm_scrolled_wb_t scwb;
	mbtk_box_t vbox;
	vtp0_t lines;
	unsigned single_line:1;

	/* private */
	unsigned focus_no_cursor:1; /* do not redraw cursor in focus received */
	unsigned lock_del:1;        /* do not react on child (line widget) removal */
	unsigned root_set:1;
	struct { long x, y; } cursor;
} mbtk_textedit_t;

/* Create a new textedit with initial string loaded (may be NULL) */
mbtk_textedit_t *mbtk_textedit_new(mbtk_textedit_t *caller_alloced, const char *str, int single_line);


/* Append a new string at the end. Multiline (\n) strings are split into
   lines for a multiline editor; in single-line mode newlines are replaced
   by spaces. */
void mbtk_textedit_append_str(mbtk_textedit_t *textedit, const char *str);

/* Append a single line of new string at the end. The _len appends exactly
   len bytes as a single. Return 0 on success. Fails in single-line mode instead
   of adding a second line */
int mbtk_textedit_append_line(mbtk_textedit_t *textedit, const char *str, long len);

/* Allocate and insert a new line after the after_nth line. Return 0 on success.
   Always fails in single-line mode. */
int mbtk_textedit_insert_line_after(mbtk_textedit_t *textedit, long after_n, const char *str, long len);

/* Remove the yth line. If cursor was in that line, remove cursor as well.
   Return 0 on success, -1 on error (y out of range). */
int mbtk_textline_del_line(mbtk_textedit_t *textedit, long y);


/* Move the cursor as if a cursor key was pressed. Returns 0 on success
   (even if cursor position did not change) or -1 on error (no cursor) */
int mbtk_textedit_cursor_move_left(mbtk_textedit_t *textedit);
int mbtk_textedit_cursor_move_right(mbtk_textedit_t *textedit);
int mbtk_textedit_cursor_move_up(mbtk_textedit_t *textedit);
int mbtk_textedit_cursor_move_down(mbtk_textedit_t *textedit);
int mbtk_textedit_cursor_move_page_up(mbtk_textedit_t *textedit);
int mbtk_textedit_cursor_move_page_down(mbtk_textedit_t *textedit);
int mbtk_textedit_cursor_move_home(mbtk_textedit_t *textedit);
int mbtk_textedit_cursor_move_end(mbtk_textedit_t *textedit);

/* Insert or overwrite text at cursor. Return 0 on success. */
int mbtk_textedit_add_chr(mbtk_textedit_t *textedit, int chr);

/* Remove a character as if backspace or del key was pressed; merge
   lines as needed. Return 0 on success. */
int mbtk_textedit_backspace(mbtk_textedit_t *textedit);
int mbtk_textedit_del(mbtk_textedit_t *textedit);

/* "Press enter": split current line at crusor, insert a new line after,
   move cursor. Returns 0 on success. */
int mbtk_textedit_enter(mbtk_textedit_t *textedit);


/* Remove the cursor (if cursor is active); cursor pos is -1;-1 after the call */
void mbtk_textedit_cursor_remove(mbtk_textedit_t *textedit);

/* Move the cursor to closest valid cursor position to x;y (which are specified
   in widget coords if textedit->vbox) */
int mbtk_textedit_cursor_set_pixel_xy(mbtk_textedit_t *textedit, long x, long y);

/* Perform the smallest amount of scroll change so that the cursor is visible.
   Doesn't do anything if cursor is not active. */
void mbtk_textedit_auto_scroll(mbtk_textedit_t *textedit);

/* Return sub-widget: scrolled, scrollbar.vert, scrollbar.horiz */
mbtk_widget_t *mbtk_textedit_subwidget(mbtk_textedit_t *textedit, mbtk_kw_t subname);


MBTK_WIDGET_INLINES(textedit)

/*** For style code ***/
int mbtk_textedit_fallback_minsize(mbtk_widget_t *w, long *mx, long *my);
void mbtk_textedit_fallback_draw_no_frame(mbtk_widget_t *w); /* draw content but no frame */
void mbtk_textedit_fallback_draw(mbtk_widget_t *w); /* draw content and frame */


#endif
